//  ----------------------------------------------------------
//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>   // for memset

#include "3ddesk.h"
#include "faceset.hpp"
#include "util.h"

#define LINEAR_MOVE_TYPE      Movement::MOVE_TYPE_WACK
#define CYLINDER_MOVE_TYPE    Movement::MOVE_TYPE_WACK
#define VIEWMASTER_MOVE_TYPE  Movement::MOVE_TYPE_WACK
#define CAROUSEL_MOVE_TYPE    Movement::MOVE_TYPE_WACK

// ==============================================================
//  FaceSet - abstract base class 
// ==============================================================

FaceSet::FaceSet ()
{
    // defaults
    ee_speed_z = 45;
    z_offset_distance = 7.0;
    face_list = glGenLists(1);
    move_type = Movement::MOVE_TYPE_WACK;
    memset(faces, 0, sizeof(faces));
}

FaceSet::~FaceSet ()
{
    int i = 0;
    while (faces[i++]) 
        delete faces[i];
    
    glDeleteLists(face_list, 1);
}

void FaceSet::set_current_face (int n) 
{
    current_face = n;
}

int FaceSet::get_current_face(void) 
{
    return current_face;
}

int FaceSet::get_number_of_faces(void) 
{
    return number_of_faces;
}

float FaceSet::get_current_z_offset(void) 
{
    return z_offset;
}

void FaceSet::set_z_offset_distance(float zod) 
{
    z_offset_distance = zod;
}

void FaceSet::goto_next(void) 
{
    current_face++;
    if (current_face == number_of_faces)
        current_face = 0;   // cycle around

    if (goto_face_mv.at_destination()) {
        goto_face_mv.init (&goto_face_val, 
                           goto_face_val, goto_face_val - per_face_change, 
                           face_change_speed, move_type);
    } else {
        goto_face_mv.init (&goto_face_val, 
                           goto_face_val, goto_face_mv.destination() - per_face_change, 
                           face_change_speed, move_type);
    }

    //oscillation.init(&oscillation_value, 0, 2, 
    //                 face_change_speed + 80,
    //                 Movement::MOVE_TYPE_SINE);

}

void FaceSet::goto_previous(void) 
{
    if (current_face == 0)
        current_face = number_of_faces - 1;  // cycle around
    else
        current_face--;
        
    if (goto_face_mv.at_destination()) {
        goto_face_mv.init (&goto_face_val, 
                           goto_face_val, goto_face_val + per_face_change, 
                           face_change_speed, move_type);
    } else {
        goto_face_mv.init (&goto_face_val, 
                           goto_face_val, goto_face_mv.destination() + per_face_change, 
                           face_change_speed, move_type);
    }

    //oscillation.init(&oscillation_value, 0, 2, 
    //                 face_change_speed + 80,
    //                 Movement::MOVE_TYPE_SINE);
}


// this goto face assumes a circular arrangement
// overload for other
float FaceSet::goto_face(int newface) 
{
    if (newface == current_face
        || newface < 0
        || newface >= number_of_faces)
        return 0;
    

    float new_rot;

    // eeek!  figure out the shortest distance to the new face... there
    // must be a more elegant way!  but for now this works...
    if (newface > current_face) {
        if ( (newface - current_face) <
             ((current_face + number_of_faces) - newface) )
        { 
            // normal case 
            new_rot = -per_face_change * (newface - current_face);
        } else {
            // "wrapped"
            new_rot =  per_face_change * ((current_face + number_of_faces) - newface);
        }
    } else {
        // newface < current_face

        if ( (current_face - newface) >
             ((newface + number_of_faces) - current_face) )
        {
            // "wrapped"
            new_rot = -per_face_change * ((newface + number_of_faces) - current_face);
        } else {
            // normal case
            new_rot =  per_face_change * (current_face - newface);
        }
    }

    if (goto_face_mv.at_destination())
        goto_face_mv.init (&goto_face_val, goto_face_val, 
                           goto_face_val + new_rot,
                           face_change_speed, move_type);
    else
        goto_face_mv.init (&goto_face_val, goto_face_val, 
                           goto_face_mv.destination() + new_rot,
                           face_change_speed, move_type);

    current_face = newface; 
    return face_change_speed;
}


void FaceSet::goto_random_face(void)
{
    int which_one;

    do {
        which_one = (int)floor ( get_random (0, number_of_faces - .1) );
    } while (which_one == current_face);

    //msgout (DEBUG, "picked random %u\n", which_one);
    goto_face(which_one);
}


void FaceSet::entry_movement_start(void) 
{
    z_off_mv.init (&z_offset, 
                   z_offset_close, get_distant_z_offset(), 
                   ee_speed_z, 
                   Movement::MOVE_TYPE_SMOOTH);
}

void FaceSet::exit_movement_start(void) 
{
    z_off_mv.init(&z_offset, 
                  z_offset, z_offset_close, 
                  ee_speed_z, Movement::MOVE_TYPE_WACK);
}


int FaceSet::is_entry_movement_done(void) 
{
    return z_off_mv.at_destination();
}

int FaceSet::is_exit_movement_done(void) 
{
    return z_off_mv.at_destination();
}

void FaceSet::entry_exit_move(void) 
{
    z_off_mv.change_a_bit();
}


extern GLint digit_text[MAX_FACES];
extern GLint texture[MAX_FACES];

void FaceSet::render()
{
    glCallList(face_list);
}

void FaceSet::make_display_list(void)
{
    int i;
    int use_texture = 1;

    glMatrixMode(GL_MODELVIEW);

    glNewList(face_list,GL_COMPILE);

    for (i = 0; i < number_of_faces; i++) {
    
        use_texture = 1;

        if (texture[i] != -1) {
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, texture[i]);   // 2d texture (x and y size)
            glColor4f(1, 1, 1, 1);
        } else {
            // no screenshot so show the digit corresponding to
            // the screen number if we have it
            if (digit_text[i] != -1) {
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, digit_text[i]);
                glColor4f(.5, .5, .5, 1);
            } else {
                glDisable(GL_TEXTURE_2D);
                use_texture = 0;
                glColor4f(.2, .2, .2, .50);
            }
        }
        faces[i]->render(use_texture);

    }
    glEndList();
}


// ==============================================================
//  CarouselFaceSet
// ==============================================================

CarouselFaceSet::CarouselFaceSet(int number, int face, float fw, float fh)
{
    X_Rot = 0;

    // must match z speed!
    speed_x = ee_speed_z;

    x_rot_final = 15.0;

    //memset (faces, 0, sizeof(faces));
        
    number_of_faces = number;
    current_face = face;
    face_width = fw;
    face_height = fh;
                
    side_angle = 360.0 / (float)number_of_faces;

    per_face_change = side_angle;

    goto_face_val = 360 - (side_angle * current_face);

    float circum = face_width * 1.5 * number_of_faces;
    radius = circum / (2.0 * PI);



    float my_angle = ((FOV / 2.0) * PI) / 180.0;
    
    z_offset_close = z_offset = -(float)((face_height/2.0) / tan(my_angle) + 
                                         get_base_z_offset());
    

    int i;
    float half_face_height = face_height / 2.0;
    float half_face_width = face_width / 2.0;

    Face *f;

    Point tl,tr,bl,br;

    for (i = 0; i < number_of_faces; i++) {
        
        
        tl.set (-half_face_width,  half_face_height, 0);
        tr.set ( half_face_width,  half_face_height, 0);
        bl.set (-half_face_width, -half_face_height, 0);
        br.set ( half_face_width, -half_face_height, 0);

        translate (tl.v, 0, 0, radius);
        rotate_y (tl.v, side_angle * i);

        translate (tr.v, 0, 0, radius);
        rotate_y (tr.v, side_angle * i);

        translate (bl.v, 0, 0, radius);
        rotate_y (bl.v, side_angle * i);

        translate (br.v, 0, 0, radius);
        rotate_y (br.v, side_angle * i);

        //msgout (DEBUG"[%i] = %f, %f, %f\n", i, tl.v[0], tl.v[1], tl.v[2]);

        f = new Face (tl, tr, bl, br);
        if (!f) {
            msgout (ERROR, "out of memory\n");
            exit (-1);
        }

#if 0
        msgout(DEBUG, "[%i] X tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[0], f->topright.v[0], f->bottomright.v[0]);
        msgout(DEBUG, "[%i] Y tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[1], f->topright.v[1], f->bottomright.v[1]);
        msgout(DEBUG, "[%i] Z tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[2], f->topright.v[2], f->bottomright.v[2]);

        msgout(DEBUG, "[%i] normal = %f, %f, %f\n\n", i, 
                f->normal.v[0], f->normal.v[1], f->normal.v[2]);
#endif

        faces[i] = f;
    }

    make_display_list();

    goto_face_mv.disable();
}


void CarouselFaceSet::set_current_face (int n) 
{
    current_face = n;
    goto_face_val = 360 - (side_angle * current_face);
    goto_face_mv.disable();
}

float CarouselFaceSet::get_distant_z_offset(void) 
{
    return -(z_offset_distance + get_base_z_offset());
}

float CarouselFaceSet::get_base_z_offset(void) 
{
    return (radius);
}


void CarouselFaceSet::do_movement(void) 
{
    if (!goto_face_mv.at_destination())
        goto_face_mv.change_a_bit ();
    else 
        clamp_degrees(&goto_face_val);

}

void CarouselFaceSet::do_position_GL(void) 
{
    glTranslatef(0.0f,0.0f, z_offset);

    glRotatef(X_Rot, 1.0f, 0.0f, 0.0f);

    //if (!oscillation.at_destination()) {
    //    oscillation.change_a_bit();
    //    //glRotatef(oscillation_value, 1.0, 0.0, 0.0);
    //    glRotatef(oscillation_value, 0.0, 1.0, 0.0);
    //}


    glRotatef(goto_face_val, 0.0f, 1.0f, 0.0f);
}


void CarouselFaceSet::entry_movement_start(void) 
{
    x_rot_mv.init (&X_Rot,
                   0.0, x_rot_final,
                   speed_x,
                   Movement::MOVE_TYPE_SMOOTH);
    FaceSet::entry_movement_start();
}

void CarouselFaceSet::exit_movement_start(void) 
{
    x_rot_mv.init(&X_Rot,
                  X_Rot, 0.0,
                  speed_x,
                  Movement::MOVE_TYPE_WACK);
    FaceSet::exit_movement_start();
}

void CarouselFaceSet::entry_exit_move(void) 
{
    x_rot_mv.change_a_bit();
    FaceSet::entry_exit_move();
}


// ==============================================================
//  CylinderFaceSet
// ==============================================================


CylinderFaceSet::CylinderFaceSet(int number, int face, float fw, float fh)
{
    z_offset_distance = 5.0;

    // x speed must match z speed!
    speed_x = ee_speed_z;
    x_rot_final = 15.0;

    memset (faces, 0, sizeof(faces));
        
    number_of_faces = number;
    current_face = face;
    face_width = fw;
    face_height = fh;
                
    side_angle = 360.0 / (float)number_of_faces;
    per_face_change = side_angle;

    goto_face_val = 360 - (side_angle * current_face);


    float my_angle = ((FOV / 2.0) * PI) / 180.0;
    
    z_offset_close = z_offset = -(float)((face_height/2.0) / tan(my_angle) + 
                                         get_base_z_offset());
    

    int i;
    float half_face_height = face_height / 2.0;
    float half_face_width = face_width / 2.0;

    Face *f;
    Point tl,tr,bl,br;
    
    if (number_of_faces == 2) {

        // 2 faces really messes up becase the faces sandwhich
        // together and you can't see one of the textures so special
        // case it and move them apart

        tl.set (-half_face_width,  half_face_height, two_face_gap);
        tr.set ( half_face_width,  half_face_height, two_face_gap);
        bl.set (-half_face_width, -half_face_height, two_face_gap);
        br.set ( half_face_width, -half_face_height, two_face_gap);

        f = new Face (tl, tr, bl, br);
        if (!f) {
            msgout(ERROR, "out of memory\n");
            exit (-1);
        }

        faces[0] = f;

        tl.set ( half_face_width,  half_face_height, -two_face_gap);
        tr.set (-half_face_width,  half_face_height, -two_face_gap);
        bl.set ( half_face_width, -half_face_height, -two_face_gap);
        br.set (-half_face_width, -half_face_height, -two_face_gap);

        f = new Face (tl, tr, bl, br);
        if (!f) {
            msgout(ERROR, "out of memory\n");
            exit (-1);
        }

        faces[1] = f;

    } else {
        polypoints_t pp;

        calculate_polypoints (number_of_faces, face_width, &pp);
        for (i = 0; i < number_of_faces; i++) {
            // translate the values calculated for a n-face polygon in
            // a x-y coord system to the 3d x-y-z coords.  we don't
            // need y because the up and down (y-axis) is always
            // either half_face_height or -half_face_height
            
            //msgout(DEBUG, "%u:: x=%lf    y=%lf\n", i, pp.x[i], pp.y[i]);
            
            tl.set( pp.x[i],  half_face_height, -pp.y[i]);
            bl.set( pp.x[i], -half_face_height, -pp.y[i]);
            
            if (i == number_of_faces - 1) {
                tr.set( pp.x[0],  half_face_height, -pp.y[0]);
                br.set( pp.x[0], -half_face_height, -pp.y[0]);
            } else {
                tr.set( pp.x[i+1],  half_face_height, -pp.y[i+1]);
                br.set( pp.x[i+1], -half_face_height, -pp.y[i+1]);
            }
            
            f = new Face (tl,tr,bl,br);
            if (!f) {
                msgout(ERROR, "out of memory\n");
                exit (-1);
            }
            
            faces[i] = f;
        }
    }

    make_display_list();

    goto_face_mv.disable();
}

void CylinderFaceSet::set_current_face (int n) 
{
    current_face = n;
    goto_face_val = 360 - (side_angle * current_face);
    goto_face_mv.disable();
}

float CylinderFaceSet::get_distant_z_offset(void) 
{
    return -(z_offset_distance + get_base_z_offset());
}

float CylinderFaceSet::get_base_z_offset(void) 
{
    if (number_of_faces == 2) {
        return two_face_gap;
    } else {
        return ((face_width/2.0) / tan (deg_to_rad(side_angle/2.0)));
    }
}


void CylinderFaceSet::do_movement(void) 
{
    if (!goto_face_mv.at_destination())
        goto_face_mv.change_a_bit ();
    else
        clamp_degrees(&goto_face_val);
}

void CylinderFaceSet::do_position_GL(void) 
{
    glTranslatef(0.0f,0.0f, z_offset);

    glRotatef(X_Rot, 1.0f, 0.0f, 0.0f);

    glRotatef(goto_face_val, 0.0f, 1.0f, 0.0f);
}

void CylinderFaceSet::entry_movement_start(void) 
{
    x_rot_mv.init (&X_Rot,
                   0.0, x_rot_final,
                   speed_x,
                   Movement::MOVE_TYPE_SMOOTH);
    FaceSet::entry_movement_start();
}

void CylinderFaceSet::exit_movement_start(void) 
{
    x_rot_mv.init(&X_Rot,
                  X_Rot, 0.0,
                  speed_x,
                  Movement::MOVE_TYPE_WACK);
    FaceSet::exit_movement_start();
}

void CylinderFaceSet::entry_exit_move(void) 
{
    x_rot_mv.change_a_bit();
    FaceSet::entry_exit_move();
}



// ==============================================================
//  ViewMasterFaceSet
// ==============================================================

ViewMasterFaceSet::ViewMasterFaceSet(int number, int face, float fw, float fh) 
{
    z_offset_distance = 10.0;

    between_angle = 360 / number;
    per_face_change = between_angle;

    center_distance = number;  // distance from the center

    if (number <= 3)
        center_distance = 3.5;

    memset (faces, 0, sizeof(faces));
        
    number_of_faces = number;
    current_face = face;
    face_width = fw;
    face_height = fh;
                
    float my_angle = ((FOV / 2.0) * PI) / 180.0;
    
    z_offset_close = z_offset = -(float)((face_height/2.0) / tan(my_angle) + 
                                         get_base_z_offset());
    
    Face *f;
    Point tl,tr,bl,br;
    int i;
    float angle;
    float center_x, center_y;
    float x1, y1, x2, y2, x3, y3, x4, y4;

    float r = sqrt ((face_height*face_height) + 
                    (face_width*face_width)) / 2.0;
    
    float angle_l, angle_s, hangle_l;

    for (i = 0; i < number_of_faces; i++) {
        angle = between_angle * i;

        center_x = polar_to_x (center_distance, angle);
        center_y = polar_to_y (center_distance, angle);

        hangle_l = rad_to_deg(atan(face_width/face_height));
        angle_l = hangle_l*2.0;
        angle_s = 180 - angle_l;
        

        x1 = polar_to_x(r, 90+angle+hangle_l) + center_x;
        y1 = polar_to_y(r, 90+angle+hangle_l) + center_y;
        x2 = polar_to_x(r, 90+angle-hangle_l) + center_x;
        y2 = polar_to_y(r, 90+angle-hangle_l) + center_y;
        x3 = polar_to_x(r, 90+angle+hangle_l+angle_s) + center_x;
        y3 = polar_to_y(r, 90+angle+hangle_l+angle_s) + center_y;
        x4 = polar_to_x(r, 90+angle+hangle_l+angle_s+angle_l) + center_x;
        y4 = polar_to_y(r, 90+angle+hangle_l+angle_s+angle_l) + center_y;

        tl.set(x1, y1, 0);
        tr.set(x2, y2, 0);
        bl.set(x3, y3, 0);
        br.set(x4, y4, 0);

#if 0
        if (i == 0) {
            msgout(DEBUG, "hangle_l=%f\n", hangle_l);
            msgout(DEBUG, "cx=%f  cy=%f\n", center_x, center_y);
            msgout(DEBUG, "r=%f\n", r);
            msgout(DEBUG, "x=%f  y=%f\n", x1, y1);
            msgout(DEBUG, "x=%f  y=%f\n", x2, y2);
            msgout(DEBUG, "x=%f  y=%f\n", x3, y3);
            msgout(DEBUG, "x=%f  y=%f\n", x4, y4);
        }
#endif

        f = new Face (tl,tr,bl,br);
        if (!f) {
            msgout(ERROR, "out of memory\n");
            exit (-1);
        }

        faces[i] = f;
    }

    make_display_list();

    goto_face_val = -current_face * between_angle;

    goto_face_mv.disable();
}


void ViewMasterFaceSet::set_current_face (int n) 
{
    current_face = n;
    goto_face_val = -current_face * between_angle;
    goto_face_mv.disable();
}

float ViewMasterFaceSet::get_distant_z_offset(void)
{
    return -(z_offset_distance);
}


float ViewMasterFaceSet::get_base_z_offset(void)
{ 
    return 0.0;
}


void ViewMasterFaceSet::do_position_GL(void) 
{
    glTranslatef(0.0f,0.0f, z_offset);
    glTranslatef(-center_distance, 0.0f, 0.0f);
    glRotatef(goto_face_val, 0.0f, 0.0f, 1.0f);
}

void ViewMasterFaceSet::do_movement(void) 
{
    if (!goto_face_mv.at_destination())
        goto_face_mv.change_a_bit();
    else
        clamp_degrees (&goto_face_val);
}


// ==============================================================
//  PriceIsRightFaceSet
// ==============================================================
PriceIsRightFaceSet::PriceIsRightFaceSet(int number, int face, float fw, float fh)
{
    memset (faces, 0, sizeof(faces));
        
    number_of_faces = number;
    current_face = face;
    face_width = fw;
    face_height = fh;
                
    side_angle = 360.0 / (float)number_of_faces;
    per_face_change = -side_angle;

    goto_face_val = 360 + (side_angle * current_face);


    float circum = face_height * 1.8 * number_of_faces;
    radius = circum / (2.0 * PI);



    float my_angle = ((FOV / 2.0) * PI) / 180.0;
    
    z_offset_close = z_offset = -(float)((face_height/2.0) / tan(my_angle) + 
                                         get_base_z_offset());
    

    int i;
    float half_face_height = face_height / 2.0;
    float half_face_width = face_width / 2.0;

    Face *f;
    Point tl,tr,bl,br;
    
    for (i = 0; i < number_of_faces; i++) {
        
        tl.set (-half_face_width,  half_face_height, 0);
        tr.set ( half_face_width,  half_face_height, 0);
        bl.set (-half_face_width, -half_face_height, 0);
        br.set ( half_face_width, -half_face_height, 0);

        translate (tl.v, 0, 0, radius);
        rotate_x (tl.v, -side_angle * i);

        translate (tr.v, 0, 0, radius);
        rotate_x (tr.v, -side_angle * i);

        translate (bl.v, 0, 0, radius);
        rotate_x (bl.v, -side_angle * i);

        translate (br.v, 0, 0, radius);
        rotate_x (br.v, -side_angle * i);

        //msgout(DEBUG, "[%i] = %f, %f, %f\n", i, tl.v[0], tl.v[1], tl.v[2]);

        f = new Face (tl, tr, bl, br);
        if (!f) {
            msgout (ERROR, "out of memory\n");
            exit (-1);
        }

#if 0
        msgout(DEBUG, "[%i] X tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[0], f->topright.v[0], f->bottomright.v[0]);
        msgout(DEBUG, "[%i] Y tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[1], f->topright.v[1], f->bottomright.v[1]);
        msgout(DEBUG, "[%i] Z tl=%f, tr=%f, br=%f\n", i, 
                f->topleft.v[2], f->topright.v[2], f->bottomright.v[2]);

        msgout(DEBUG, "[%i] normal = %f, %f, %f\n\n", i, 
                f->normal.v[0], f->normal.v[1], f->normal.v[2]);
#endif

        faces[i] = f;
    }

    make_display_list();

    goto_face_mv.disable();
}


void PriceIsRightFaceSet::set_current_face (int n) 
{
    current_face = n;
    goto_face_val = 360 + (side_angle * current_face);
    goto_face_mv.disable();
}

float PriceIsRightFaceSet::get_distant_z_offset(void) 
{
    return -(z_offset_distance + get_base_z_offset());
}

float PriceIsRightFaceSet::get_base_z_offset(void) 
{
    return (radius ); //(face_width/2.0) / tan (deg_to_rad(side_angle/2.0)));
}


void PriceIsRightFaceSet::do_movement(void) 
{
    if (!goto_face_mv.at_destination())
        goto_face_mv.change_a_bit ();
    else
        clamp_degrees(&goto_face_val);
}

void PriceIsRightFaceSet::do_position_GL(void) 
{
    glTranslatef(0.0f,0.0f, z_offset);

    glRotatef(goto_face_val, 1.0f, 0.0f, 0.0f);
}


// ==============================================================
//  LinearFaceSet
// ==============================================================

LinearFaceSet::LinearFaceSet(int number, int face, float fw, float fh) 
{
    z_offset_distance = 10.0;
    between = 2.0;

    spin_z_rot = 0;
    ee_speed_z = 60;

    memset (faces, 0, sizeof(faces));
        
    number_of_faces = number;
    current_face = face;
    face_width = fw;
    face_height = fh;

    float my_angle = ((FOV / 2.0) * PI) / 180.0;
    
    z_offset_close = z_offset = -(float)((face_height/2.0) / tan(my_angle) + 
                                         get_base_z_offset());
    
    //msgout(DEBUG, "z_offset_close = %lf\n", z_offset_close);
    

    Face *f;
    Point tl,tr,bl,br;
    int i;
    float x;
    float half_face_height = face_height / 2.0;
    float half_face_width = face_width / 2.0;

    X_Trans = -(half_face_width + (current_face * (face_width + between)));

    for (i = 0; i < number_of_faces; i++) {
        x = i * (face_width + between);

        tl.set(x,               half_face_height, 0);
        tr.set(x + face_width,  half_face_height, 0);
        bl.set(x,              -half_face_height, 0);
        br.set(x + face_width, -half_face_height, 0);

        f = new Face (tl,tr,bl,br);
        if (!f) {
            msgout(ERROR, "out of memory\n");
            exit (-1);
        }

        faces[i] = f;
    }

    make_display_list();

    x_trans_mv.disable();
    spin_mv.disable();
}


void LinearFaceSet::set_current_face (int n) 
{
    float half_face_width = face_width / 2.0;

    current_face = n;
    X_Trans = -(half_face_width + (current_face * (face_width + between)));
    x_trans_mv.disable();
}

float LinearFaceSet::get_distant_z_offset(void)
{
    return -(z_offset_distance);
}


float LinearFaceSet::get_base_z_offset(void)
{ 
    return 0.0;
}

void LinearFaceSet::goto_previous(void) 
{
    if (current_face != 0)
        current_face--;
    else 
        return;

    if (x_trans_mv.at_destination()) {
        //if ((X_Trans + (face_width+between)) > MIN_X_Trans)   // because its negative
        //    break;
        x_trans_mv.init (&X_Trans, X_Trans, X_Trans + (face_width+between), face_change_speed, LINEAR_MOVE_TYPE);
    } else {
        //if ((x_trans_mv.destination() + (face_width+between)) > MIN_X_Trans)   // because its negative
        //    break;
        x_trans_mv.init (&X_Trans, X_Trans, x_trans_mv.destination() + (face_width+between), face_change_speed, LINEAR_MOVE_TYPE);
    }
}

void LinearFaceSet::goto_next(void) 
{
    if (current_face != (number_of_faces - 1))
        current_face++;
    else
        return;
    
    if (x_trans_mv.at_destination()) {
        x_trans_mv.init (&X_Trans, X_Trans, X_Trans - (face_width + between), 
                         face_change_speed, LINEAR_MOVE_TYPE);
    } else {
        x_trans_mv.init (&X_Trans, X_Trans, 
                         x_trans_mv.destination() - (face_width+between), 
                         face_change_speed, LINEAR_MOVE_TYPE);
    }
}

float LinearFaceSet::goto_face(int newface) 
{
    signed int nummove;
    
    if (newface == current_face
        || newface < 0
        || newface >= number_of_faces)
        return 0;
    
    nummove = newface - current_face;
    current_face = newface; 
    
    if (x_trans_mv.at_destination()) {
        x_trans_mv.init (&X_Trans, X_Trans, 
                         X_Trans - ((float)nummove * (face_width + between)), 
                         face_change_speed, LINEAR_MOVE_TYPE);
    } else {
        x_trans_mv.init (&X_Trans, X_Trans, 
                         x_trans_mv.destination() - ((float)nummove * (face_width + between)), 
                         face_change_speed, LINEAR_MOVE_TYPE);
    }
    return face_change_speed;
}

void LinearFaceSet::do_position_GL(void) 
{
    if (spin_z_rot)
        glRotatef(spin_z_rot,0.0f,0.0f,1.0f);

    glTranslatef(0.0f,0.0f, z_offset);
    glTranslatef(X_Trans, 0.0f, 0.0f);
}

void LinearFaceSet::do_movement(void) 
{
    if (!x_trans_mv.at_destination()) {
        x_trans_mv.change_a_bit();
    }
}

void LinearFaceSet::entry_exit_move(void) 
{
    spin_mv.change_a_bit();
    FaceSet::entry_exit_move();
}

void LinearFaceSet::exit_movement_start(void) 
{
    spin_mv.init (&spin_z_rot, 
                  0.0, 360.0, 
                  ee_speed_z, 
                  Movement::MOVE_TYPE_WACK);
    FaceSet::exit_movement_start();
}

