//  ----------------------------------------------------------
//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef _FACESET_HPP
#define _FACESET_HPP

#include "3ddesk.h"
#include "face.hpp"
#include "move.hpp" // for Movement

// FIXME - use dynamic list or something
// MAX_FACES  must be defined before including this file


// ===================================================================
//  FaceSet - abstract base class 
//    This defines the interface to the derived classes below.  Allows
//    a generic pointer to "FaceSet" to be used whose calls to the
//    virtual functions are then sent to the derived class.
// ===================================================================
class FaceSet {
protected:
    int number_of_faces;
    int current_face;

    Movement::MOVE_TYPE move_type;

    float face_width;
    float face_height; 

    const static float face_change_speed = 40;

    Movement oscillation;
    float oscillation_value;

    // entry/exist z offset speed
    float ee_speed_z;

    Movement z_off_mv;

    // the current z offset changed by the above movement
    float z_offset;

    // this is the z offset that makes the current face match the screen
    // (all the edges of the quad are aligned with the edges of the
    // screen) It's calculated dynamically based on the screen dimensions
    float z_offset_close;

    // this is the space between the viewer and the face at a distance
    float z_offset_distance;

    Face *faces[MAX_FACES];

    GLuint face_list;

    Movement goto_face_mv;
    float goto_face_val;
    float per_face_change;

public:
    FaceSet ();
    virtual ~FaceSet ();

    //virtual ~FaceSet() { };
    
    virtual void goto_next(void);
    virtual void goto_previous(void);
    virtual void goto_random_face(void);
    virtual float goto_face(int newface);

    // methods that must be implemented by derived classes
    virtual float get_base_z_offset(void) = 0;
    virtual float get_distant_z_offset(void) = 0;
    virtual void do_movement(void) = 0;
    virtual void do_position_GL(void) = 0;
    virtual void render();
    virtual void make_display_list(void);

    // movement hooks for controling entry and exit movements (zoom
    // out and zoom in etc)
    virtual void entry_movement_start(void);
    virtual void entry_exit_move(void);
    virtual void exit_movement_start(void);
    virtual int is_entry_movement_done(void);
    virtual int is_exit_movement_done(void);

    // "locally" defined functions
    virtual void set_current_face (int n);
    int get_current_face(void);
    int get_number_of_faces(void);
    float get_current_z_offset(void);
    void set_z_offset_distance(float zod);

};



class CarouselFaceSet : public FaceSet {
private:
    const static float two_face_gap = 3;

    Movement x_rot_mv;

    float side_angle;

    float speed_x;
    float x_rot_final;
    float X_Rot;

    float radius;

public:
    CarouselFaceSet(int number, int face, float fw, float fh);

    void set_current_face (int n);
    float get_base_z_offset(void); 
    float get_distant_z_offset(void); 
    void do_movement(void);
    void do_position_GL(void);

    // Entry/Exit movement functions
    void entry_movement_start(void);
    void exit_movement_start(void);
    void entry_exit_move(void);
};


class CylinderFaceSet : public FaceSet {
private:
    const static float two_face_gap = 3;

    float side_angle;

    Movement x_rot_mv;

    float X_Rot;

    float speed_x;
    float x_rot_final;

public:
    CylinderFaceSet(int number, int face, float fw, float fh);

    void set_current_face (int n);
    float get_base_z_offset(void); 
    float get_distant_z_offset(void); 
    void do_movement(void);
    void do_position_GL(void);

    // Entry/Exit movement functions
    void entry_movement_start(void);
    void exit_movement_start(void);
    void entry_exit_move(void);

};



class ViewMasterFaceSet : public FaceSet {
private:
    float between_angle;
    float center_distance;

public:
    ViewMasterFaceSet(int number, int face, float fw, float fh);

    void set_current_face (int n);
    float get_base_z_offset(void); 
    float get_distant_z_offset(void); 
    void do_movement(void);
    void do_position_GL(void);
};



class PriceIsRightFaceSet : public FaceSet {
private:
    const static float two_face_gap = 3;

    float side_angle;

    float radius;
public:
    PriceIsRightFaceSet(int number, int face, float fw, float fh);

    void set_current_face (int n);
    float get_base_z_offset(void); 
    float get_distant_z_offset(void); 
    void do_movement(void);
    void do_position_GL(void);
};



class LinearFaceSet : public FaceSet {
private:

    float between;

    Movement x_trans_mv;
    float X_Trans;

    Movement spin_mv;
    float spin_z_rot;

public:
    LinearFaceSet(int number, int face, float fw, float fh);

    void set_current_face (int n);
    float get_base_z_offset(void); 
    float get_distant_z_offset(void); 
    void goto_next(void);
    void goto_previous(void);
    float goto_face(int newface);
    void do_movement(void);
    void do_position_GL(void);

    // Entry/Exit movement functions which add spinning on exit
    void entry_exit_move(void);
    void exit_movement_start(void);

};


#endif // _FACESET_HPP
