//
//  Copyright (C) 2002 Brad Wasson <bard@systemtoolbox.com>
//
//  This file is part of 3ddesktop.
//
//  3ddesktop is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  3ddesktop is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with 3ddesktop; see the file COPYING.   If not, write to
//  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <GL/gl.h>   // OpenGL itself.
#include <GL/glu.h>  // GLU support library.
#include <GL/glx.h>

#include <X11/extensions/xf86vmode.h>
#include <Imlib2.h>

#include "win.hpp"

#include "3ddesk.h" // for FOV

/* attributes for a single buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 16,
    None
};

/* attributes for a double buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListDbl[] = { GLX_RGBA, GLX_DOUBLEBUFFER, 
    GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 16,
    None 
};


unsigned int GLWindow::get_width()
{ return width; }

unsigned int GLWindow::get_height()
{ return height; }

unsigned int GLWindow::get_screen_width()
{ return screen_width; }

unsigned int GLWindow::get_screen_height()
{ return screen_height; }


Display *GLWindow::get_display()
{ return display; }

Window GLWindow::get_window()
{ return win; }


extern unsigned int texture_size;

int GLWindow::open_display(int verb) 
{
    verbose = verb;

    display = XOpenDisplay(0);
    if (display == NULL) {
        msgout (ERROR, "Can't open display\n");
        return -1;
    }
    
    screen = DefaultScreen(display);

    screen_width = DisplayWidth(display, screen);
    screen_height = DisplayHeight(display, screen);

    screenshot_data = new unsigned char [texture_size * texture_size * 4];
    if (screenshot_data == NULL) {
        msgout (ERROR, "out of memory\n");
        end_program(-1);
    }

    Visual *vis = DefaultVisual(display, screen);
    Colormap cm = DefaultColormap(display, screen);

    imlib_context = imlib_context_new();
    imlib_context_set_display(display);
    imlib_context_set_visual(vis);
    imlib_context_set_colormap(cm);
    imlib_context_set_drawable(RootWindow(display, screen));
    imlib_context_set_anti_alias(1);
    imlib_context_set_blend(0);

    imgfinal = imlib_create_image(texture_size, texture_size);
    img = imlib_create_image(screen_width, screen_height);

    return 0;
}


int GLWindow::create_GL_window(char *title, Bool fs)
{
    Colormap cmap;
    int dpyWidth, dpyHeight;
    int i;
    int glxMajorVersion, glxMinorVersion;
    int vidModeMajorVersion, vidModeMinorVersion;
    XF86VidModeModeInfo **modes;
    int modeNum;
    int bestMode;
    Atom wmDelete;
    Window winDummy;
    unsigned int borderDummy;
    
    static int first_time = 1;

    /* set best mode to current */
    bestMode = 0;

    fullscreen = fs;

    if (fullscreen) {
        width = screen_width;
        height = screen_height;
        msgout (DEBUG, "width %d x %d\n", width, height);
    } else {
        width = 600;
        height = 480;
    }


    XF86VidModeQueryVersion(display, &vidModeMajorVersion,
                            &vidModeMinorVersion);

    msgout(DEBUG, "XF86VidModeExtension-Version %d.%d\n", vidModeMajorVersion,
           vidModeMinorVersion);

    XF86VidModeGetAllModeLines(display, screen, &modeNum, &modes);
    /* save desktop-resolution before switching modes */
    deskMode = *modes[0];
    /* look for mode with requested resolution */
    for (i = 0; i < modeNum; i++)
    {
        if ((modes[i]->hdisplay == width) && (modes[i]->vdisplay == height))
        {
            bestMode = i;
        }
    }

    /* get an appropriate visual */
    vi = glXChooseVisual(display, screen, attrListDbl);
    if (vi == NULL)
    {
        vi = glXChooseVisual(display, screen, attrListSgl);
        msgout(DEBUG, "Only Singlebuffered Visual!\n");
    }
    else
    {
        msgout(DEBUG, "Got Doublebuffered Visual!\n");
    }
    glXQueryVersion(display, &glxMajorVersion, &glxMinorVersion);
    
        msgout(DEBUG, "glX-Version %d.%d\n", glxMajorVersion, glxMinorVersion);
    


    if (first_time) {
        /* create a GLX context */
        msgout (DEBUG, "create context\n");
        ctx = glXCreateContext(display, vi, 0, GL_TRUE);
        
        first_time = 0;
    }

    /* create a color map */
    cmap = XCreateColormap(display, RootWindow(display, vi->screen),
                           vi->visual, AllocNone);
    
    
    attr.colormap = cmap;
    attr.border_pixel = 0;

    if (fullscreen)
    {
        fullscreenMode = *modes[bestMode];

        XF86VidModeSwitchToMode(display, screen, modes[bestMode]);
        XF86VidModeSetViewPort(display, screen, 0, 0);
        dpyWidth = modes[bestMode]->hdisplay;
        dpyHeight = modes[bestMode]->vdisplay;

        msgout(DEBUG, "Resolution %dx%d\n", dpyWidth, dpyHeight);

        XFree(modes);
    
        /* create a fullscreen window */
        attr.override_redirect = True;
        attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        win = XCreateWindow(display, RootWindow(display, vi->screen),
                                   0, 0, dpyWidth, dpyHeight, 0, vi->depth, InputOutput, vi->visual,
                                   CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
                                   &attr);

        // this code fragment from xscreensaver - for making a blank mouse pointer
        XColor black;
        black.red = black.green = black.blue = 0;

        Pixmap bit;
        bit = XCreatePixmapFromBitmapData (display, win,
                                           "\000", 1, 1,
                                           0, //BlackPixelOfScreen (screen),
                                           0, //BlackPixelOfScreen (screen),
                                           1);
        empty_cursor = XCreatePixmapCursor (display, bit, bit, &black, &black,
                                             0, 0);
        XFreePixmap (display, bit);
        // -----

        //XUndefineCursor (display, win);

        //XDefineCursor (display, win, empty_cursor);

        //XWarpPointer(display, None, win, 0, 0, 0, 0, 0, 0);
        
        XMapRaised(display, win);
        XGrabKeyboard(display, win, True, GrabModeAsync,
                      GrabModeAsync, CurrentTime);
        XGrabPointer(display, win, True, ButtonPressMask,
                     GrabModeAsync, GrabModeAsync, win, empty_cursor, CurrentTime);
    }
    else
    {
        /* create a window in window mode*/
        attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        win = XCreateWindow(display, RootWindow(display, vi->screen),
                                   0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
                                   CWBorderPixel | CWColormap | CWEventMask, &attr);
        /* only set window title and handle wm_delete_events if in windowed mode */
        wmDelete = XInternAtom(display, "WM_DELETE_WINDOW", True);
        XSetWMProtocols(display, win, &wmDelete, 1);
        XSetStandardProperties(display, win, title,
                               title, None, NULL, 0, NULL);
        XMapRaised(display, win);
    }       
    /* connect the glx-context to the window */
    glXMakeCurrent(display, win, ctx);
    XGetGeometry(display, win, &winDummy, &x, &y,
                 (unsigned int *)&width, (unsigned int *)&height, 
                 &borderDummy, &depth);

    msgout(DEBUG, "Depth %d\n", depth);

    if (glXIsDirect(display, ctx)) 
        msgout(DEBUG, "Congrats, you have Direct Rendering!\n");
    else
        msgout(DEBUG, "Sorry, no Direct Rendering possible!\n");
    
    return 1;
}

int GLWindow::kill_GL_window ()
{
    msgout (DEBUG, "kill window\n");

    XDestroyWindow(display, win);

    XSync(display, 0);

    /* switch back to original desktop resolution if we were in fs */
    if (fullscreen)
    {
        XF86VidModeSwitchToMode(display, screen, &deskMode);
        XF86VidModeSetViewPort(display, screen, 0, 0);
    }
    return 0;
}

int GLWindow::unhide_GL_window()
{
    if (fullscreen) {
        XF86VidModeSwitchToMode(display, screen, &fullscreenMode);
        XF86VidModeSetViewPort(display, screen, 0, 0);
    }

    XMapRaised (display, win);
    
    if (fullscreen) {
        XGrabKeyboard(display, win, True, GrabModeAsync,
                      GrabModeAsync, CurrentTime);
        XGrabPointer(display, win, True, ButtonPressMask,
                     GrabModeAsync, GrabModeAsync, win, empty_cursor, CurrentTime);
    }


    XSync(display, 0);
    return 0;
}

int GLWindow::hide_GL_window()
{
    XUnmapWindow (display, win);
    
    XSync(display, 0);

    if (fullscreen)
    {
        XF86VidModeSwitchToMode(display, screen, &deskMode);
        XF86VidModeSetViewPort(display, screen, 0, 0);
    }
    return 0;
}

void
GLWindow::resize_scene()
{
    resize_scene(width, height);
}

void
GLWindow::resize_scene(int w, int h)
{
    // Let's not core dump, no matter what.
    if (h == 0)
        h = 1;

    width = w;
    height = h;

    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(FOV, (GLfloat)width/(GLfloat)height,0.1f,100.0f);

    glMatrixMode(GL_MODELVIEW);

} // END resize_scene



int GLWindow::grab_screenshot_data()
{
    /* set the display , visual, colormap and drawable we are using */
    imlib_context_set_image(img);
    imlib_copy_drawable_to_image((Pixmap)0, //Pixmap mask, 
                                 0, 0, //int x, int y, 
                                 screen_width, //int width, int height,
                                 screen_height,
                                 0, 0, //int destination_x, int destination_y,
                                 1 ); //char need_to_grab_x);

    imlib_context_set_image(imgfinal);
    imlib_image_clear();

    imlib_blend_image_onto_image(img, 0, 
                                 0, 0, screen_width, screen_height, 
                                 0, 0, texture_size, texture_size);

    unsigned int *tmp = imlib_image_get_data();

    int w = imlib_image_get_width();
    int h = imlib_image_get_height();

    int x, y;
    int offset;
    int img_offset;

    w = h = texture_size;
    
    for (y = 0; y < h; y++) {

        for (x = 0; x < w; x++) {

            offset = (y * w + x) * 4;
            img_offset = (y * w + x);

            screenshot_data[offset]     = (tmp[img_offset] >> 16) & 0xff;
            screenshot_data[offset + 1] = (tmp[img_offset] >> 8) & 0xff;
            screenshot_data[offset + 2] =  tmp[img_offset] & 0xff;
            screenshot_data[offset + 3] = (tmp[img_offset] >> 24) & 0xff;


        }
    }

    return 0;
}



int GLWindow::load_screenshot_data(GLuint *texture)
{
    if (*texture == (GLuint)-1)
        glGenTextures(1, texture);
    glBindTexture(GL_TEXTURE_2D, *texture);   // 2d texture (x and y size)
    
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, MIN_FILTER);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, MAX_FILTER);
    
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, texture_size, texture_size, GL_RGBA, GL_UNSIGNED_BYTE, screenshot_data);

    return 0;
}

